#! /usr/bin/env python
# -*- coding: utf-8 -*-
################################################################################
# Indigo Plugin to control a APC Power Distribution Unit (PDU)
#
# This plugin was tested with a AP9211 with an AP9606 Web/SNMP Management
# SmartSlot Card installed.
#
#       This " should " work with any APC MasterSwitch PDU
#
# Copyright (c) 2015, Nick Thomas. All rights reserved
# v0.5
#


import time
import subprocess
import urllib2
import socket


################################################################################
# 
################################################################################
class Plugin(indigo.PluginBase):
    
    ########################################
    def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):    
        indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, 
                                   pluginVersion, pluginPrefs)
        self.debug = pluginPrefs.get("showDebugInfo", False)
            
    def __del__(self):
        indigo.PluginBase.__del__(self)

    ########################################
    def startup(self):
        self.debugLog(u"startup called")
        self.versionCheck()
        if self.pluginPrefs["processTraps"] == True:
            self.snmptrapdStart()
            
    def shutdown(self):
        self.debugLog(u"shutdown called")

    def stopConcurrentThread(self):
        self.stopThread = True
        if self.pluginPrefs["processTraps"] == True:
            self.snmptrapdKill()

    ########################################
    def versionCheck(self):
        self.debugLog(u"versionCheck() called")
        pluginName = self.pluginId.split(".")[2]
        currentVersion = str(self.pluginVersion)
        versionCheckSite = r'http://159.122.4.194/~nick'
        VersionCheck_url = versionCheckSite + "/plugins/%s.html" % pluginName
        socket.setdefaulttimeout(3)
        
        try:
            latestVersion = urllib2.urlopen(VersionCheck_url).read().split('\n')[0]

            if currentVersion < latestVersion:
               self.errorLog(u"You are running %s. A newer version, %s is available." 
                             % (currentVersion, latestVersion))
        except:
            self.errorLog(u"Unable to reach the version server.")


    ########################################
    def deviceStartComm(self, dev):
        self.debugLog("Starting device: " + dev.name)
        
        # get IP, community name and outlet from props & add to opener
        community = dev.pluginProps["community"]
        pduIpAddr = dev.pluginProps["ipAddr"]
        outlet = dev.pluginProps["outlet"]
        
        self.debugLog("Community Name: " + community)
        self.debugLog("IP address: " + pduIpAddr)
        self.debugLog("Outlet: " + outlet)
        
        # get the state of the outlets
        self.readAndUpdateState(dev)

    ########################################
    def validateDeviceConfigUi(self, valuesDict, typeId, devId):
        # validate supplied values
        outletNum = int(valuesDict["outlet"])
        if outletNum < 1 or outletNum > 8: 
            self.errorLog(u"Error: Outlet \"%s\" must be between 1 & 8" %\
                           str(outletNum))
            errorDict = indigo.Dict()
            errorDict["outlet"] = "The value of this field must be between 1 & 8"
            return (False, valuesDict, errorDict)
        else:
            return True

    ###################################################

    def shellCommand(self,theCommand,shellValue):
        proc=subprocess.Popen(theCommand, 
            shell=shellValue, 
            stdin=subprocess.PIPE, 
            stdout=subprocess.PIPE, 
            stderr=subprocess.PIPE, )
        stdout_value, stderr_value = proc.communicate()
        return stdout_value, stderr_value

    ###################################################
    def snmptrapdStart(self):
        
        if not self.is_snmptrapd_running(): 

            theCommand = ("sudo /usr/sbin/snmptrapd -m PowerNet-MIB -c "
                "'/Library/Application Support/Perceptive Automation/Indigo 6"
                "/Plugins/APC PDU Control.indigoPlugin/Contents/"
                "Server Plugin/snmptrapd.conf'")

            stdout_value, stderr_value = self.shellCommand(theCommand,True)
            
            time.sleep(0.5)

            if self.is_snmptrapd_running():
                indigo.server.log(u"snmptrapd was successfully started.")
                return True
            else:
                indigo.server.log(u"snmptrapd failed to start.")
                indigo.server.log(u"Unable to Process SNMP Traps. " 
                                  "Please review the README.")
                if stderr_value != "": self.debugLog(u"%s" % (stderr_value))                    
                return False
                
        else: #snmptrapd is already running
            indigo.server.log(u"snmptrapd was already running!")
            return True

    ###################################################
    def is_snmptrapd_running(self):
        
        theCommand = ['ps', '-U','root']
        stdout_value, stderr_value = self.shellCommand(theCommand,False)

        for line in stdout_value.splitlines():

            if 'snmptrapd -m PowerNet-MIB' in line:
                # snmptrapd IS running
                return True
        
        # snmptrapd IS NOT running
        return False

    ###################################################
    def snmptrapdKill(self):

        theCommand = ['ps', '-U','root']
        stdout_value, stderr_value = self.shellCommand(theCommand,False)

        for line in stdout_value.splitlines():

            if 'snmptrapd -m PowerNet-MIB' in line:
                pid = int(line.split(None, 1)[0])

                theCommand = "sudo /bin/kill -TERM " + str(pid)
                stdout_value, stderr_value = self.shellCommand(theCommand,True)
                
                time.sleep(0.5)
                
                if not self.is_snmptrapd_running(): 
                    indigo.server.log(u"snmptrapd has been stopped (pid %s)." 
                                      % (pid))
                    return True
                else:
                    indigo.server.log(u"snmptrapd failed to stop.")
                    return False
                
        # snmptrapd was NOT running!
        return True   

    ########################################
    def actionControlDimmerRelay(self, action, dev):
        
        ###### TURN ON ######
        if action.deviceAction == indigo.kDimmerRelayAction.TurnOn:
            self.setPDUState(dev, "on")
            
        ###### TURN OFF ######
        elif action.deviceAction == indigo.kDimmerRelayAction.TurnOff:
            self.setPDUState(dev, "off")    
            
        ###### TOGGLE ######
        elif action.deviceAction == indigo.kDimmerRelayAction.Toggle:
            newOnState = not dev.onState
            if newOnState == True:
                # Device is currently off, so turn it on
                self.setPDUState(dev, "on")
            else:
                # Device is currently on, so turn it off
                self.setPDUState(dev, "off")
                        
        ###### STATUS REQUEST ######
        elif action.deviceAction == indigo.kDeviceGeneralAction.RequestStatus:
            self.readAndUpdateState(dev)


    ########################################
    def setPDUState(self, dev, state):
        # send command to PDU to change state of an outlet
        # state argument is either "on" or "off"
        
        # validate inputs
        if state == "on":
            theStateCode = " i 1"
            OnOffState = True
                
        elif state == "off":
            theStateCode = " i 2"
            OnOffState = False
            
        else:
            self.errorLog(u"Error: State must be on or off")
            return(False)

        # get IP, community name & outlet from props
        outlet = dev.pluginProps["outlet"]
        pduIpAddr = dev.pluginProps["ipAddr"]
        community = dev.pluginProps["community"]

        # build the string to send to the PDU
        theCommand = ("snmpset -m PowerNet-MIB -v 1 -c " +
                        community + " -v 1 " + pduIpAddr +
                        " sPDUOutletCtl." + str(outlet) + theStateCode)

        # Try five (5) times, pausing for 30 seconds     
        for r in range (1,6):

            stdout_value, stderr_value = self.shellCommand(theCommand,True)

            self.debugLog(u"Sending to PDU: " + theCommand)
            self.debugLog(u"stdout value: " + stdout_value)                    

            if stderr_value == '':
                # everything worked
                self.debugLog(u"Sent to PDU")
                Successful = True
                break
            else:
                # some type of error
                self.debugLog(u"Failed to send to PDU")
                self.debugLog(u"stderr value: " + stderr_value)
                Successful = False
                time.sleep(30)

        # everything worked return True, else return False        
        if Successful:
            indigo.server.log(u"Turned \"%s\" %s" % (dev.name, state))
            dev.updateStateOnServer("onOffState", OnOffState)
        else:
            indigo.server.log(u"send \"%s\" %s failed" % (dev.name, state), 
                              isError=True)
                    
            
    ########################################
    def getPDUState(self, dev):
        # request state of PDU
        # returns state of all outlets in a single string
        
        # get IP, community name & outlet from props
        outlet = dev.pluginProps["outlet"]
        pduIpAddr = dev.pluginProps["ipAddr"]
        community = dev.pluginProps["community"]
    
        theCommand = ("snmpwalk -m PowerNet-MIB -v 1 -c " + community +
                      " " + pduIpAddr + " sPDUMasterState")

        for r in range (1,6):
            # Execute command and capture the output
            stdout_value, stderr_value = self.shellCommand(theCommand,True)
            
            if stderr_value == '':
                #create list of Off/On states returned
                theStatus = stdout_value[43:-1].split() 
        
                if theStatus[int(outlet)-1] == "Off":
                    resultCode = 0
                elif theStatus[int(outlet)-1] == "On":
                    resultCode = 1
                else:
                    resultCode = 2
                break
            else:
                resultCode = 2
                if r == 1:self.errorLog(u"Error: Retrying connection to Device %s"
                                        % dev.name)
                time.sleep(5)

        # everything worked - return resultCode = 2 if not
        self.debugLog(u"Sent to PDU: " + theCommand)        
        return(resultCode)

    ########################################
    def readAndUpdateState(self, dev):
        
        # request state from the PDU & update state variable accordingly
        resultCode = self.getPDUState(dev)
        
        if resultCode == 0:
            dev.updateStateOnServer("onOffState", False)
            indigo.server.log(u'Device "%s" is off' % (dev.name))
            return(True)
        elif resultCode == 1:
            dev.updateStateOnServer("onOffState", True)
            indigo.server.log(u'Device "%s" is on' % (dev.name))
            return(True)
        else:
            self.errorLog(u'Error: Device "%s" in unknown state' % dev.name)
            return(False)
            
    ########################################            
    def confirmStatusAll(self, pluginAction):

            # This function is meant to be executed as a Action Group,
            # called from an external SNMP trap handler.
            
            # It verifies that all outlet states in Indigo are correctly 
            # displayed an corrects them if not. Needed if states are change  
            # from PDU web portal directly.

            ipList = []
            Devices = []
            for device in indigo.devices.iter("com.anyone.apcpdu"):
                id = device.id
                if device.configured:
                    ip = device.globalProps['com.anyone.apcpdu']['ipAddr']
                    outlet = device.globalProps['com.anyone.apcpdu']['outlet']
                    Devices.append([str(ip),id,str(outlet),device.onState])
                    if str(ip) not in ipList:
                       ipList.append(str(ip))

            # Just in case there is more than one PDU
            for theIP in ipList:

                theCommand = ("snmpwalk -m PowerNet-MIB -v 1 -c private "
                              + theIP + " sPDUMasterState")
                proc=subprocess.Popen(theCommand, 
                    shell=True, 
                    stdin=subprocess.PIPE, 
                    stdout=subprocess.PIPE, 
                    stderr=subprocess.PIPE, )
                stdout_value, stderr_value = proc.communicate()

                if stderr_value == '':
                    myVar = stdout_value[43:-3]
                    theList = myVar.split()

                    for device in Devices:
                        ip, id, outlet, onState = device

                        if theIP == ip:
                            if onState:
                                if theList[int(outlet)-1] == "Off":
                                    indigo.device.turnOff(id)
                            else:
                                if theList[int(outlet)-1] == "On":
                                    indigo.device.turnOn(id)                  
